from strictyaml import Validator
from strictyaml.validators import MapValidator
from strictyaml.exceptions import YAMLSerializationError, InvalidValidatorError
from .control import Control
from .blocks import Block
from collections.abc import Callable
from .builder import ValidatorBuilder
from strictyaml.representation import YAML


class DMap(MapValidator):
    def __init__(
        self,
        control: Control,
        blocks: list[Block],
        constraints: list[Callable[..., bool]] | None = None,
    ):
        assert isinstance(control, Control), "control must be of type Control"
        assert isinstance(blocks, list), "blocks must be a list of Block"
        for block in blocks:
            assert isinstance(block, Block), "all blocks must be of type Block"
            
        if constraints is not None:
            assert isinstance(constraints, list), "constraints must be a list of Callable"
            for constraint in constraints:
                assert callable(constraint), "every constraint must be callable"

        self.control = control
        self.blocks = blocks
        self.constraints = constraints

    @staticmethod
    def compile_when(when):
        if callable(when):
            return when
        return lambda raw, ctrl: bool(when)

    @staticmethod
    def compile_constraint(when):
        if callable(when):
            return when
        return lambda raw, ctrl, val: bool(when)

    def validate(self, chunk):
        chunk.expect_mapping()
        self.control.validate(chunk)
        ctrl = self.control.validated.data
        raw = chunk.whole_document
        control_validator = self.control._validator
        true_case_block = None
        # TODO: what if the user doesn't really want a control validator and only selects based on raw
        for block in self.blocks:
            if DMap.compile_when(block.when)(raw, ctrl):
                if true_case_block is None:
                    true_case_block = block
                else:
                    chunk.expecting_but_found("when evaluating DMap blocks", "multiple cases were true")

        if true_case_block is None:
            chunk.expecting_but_found("when evaluating DMap blocks", "none of the cases were true")

        final_validator = ValidatorBuilder(
            control_validator, true_case_block._validator, self.control.source
        ).validator

        self.validated = final_validator(chunk)
        val = self.validated.data

        if self.constraints:
            for constraint in self.constraints:
                if not DMap.compile_constraint(constraint)(raw, ctrl, val):
                    chunk.expecting_but_found("when evaluating DMap constraints", "constraints not fulfilled")
        if true_case_block.constraints:
            for constraint in true_case_block.constraints:
                if not DMap.compile_constraint(constraint)(raw, ctrl, val):
                    chunk.expecting_but_found("when evaluating DMap case constraints", "constraints not fulfilled")

    def to_yaml(self, data):
        self._should_be_mapping(data)

        for block in self.blocks:
            try:
                final_validator = ValidatorBuilder(
                    self.control._validator, block._validator, self.control.source
                ).validator
                return final_validator.to_yaml(data)
            except (YAMLSerializationError, KeyError):
                continue
                
        raise YAMLSerializationError("None of the DMap cases successfully serialized the data")

    def __repr__(self):
        return "DMap({0}, {1}{2})".format(
            repr(self.control),
            repr(self.blocks),
            ", constraints={0}".format(repr(self.constraints)) if self.constraints else "",
        )



from strictyaml import Validator
from strictyaml.validators import MapValidator
from strictyaml import Map, MapCombined
import copy
from .utils import unpack, ensure_validator_dict


class ValidatorBuilder:
    def __init__(
        self,
        control_validator: Validator,
        case_validator: Validator,
        control_source: tuple[str] | str | None = None,
    ):
        self.control_source = control_source
        self.control_validator = control_validator
        self.case_validator = case_validator
        self.validator = self._build()

    def merge_recursive(self, control_validator, case_validator):
        control_validator = unpack(control_validator)
        if not hasattr(control_validator, '_validator') or not isinstance(control_validator._validator, dict):
            return
        if not hasattr(case_validator, '_validator') or not isinstance(case_validator._validator, dict):
            return

        for key, val in control_validator._validator.items():
            val_unpacked = unpack(val)
            if isinstance(val_unpacked, MapValidator):
                if key not in case_validator._validator:
                    case_validator._validator[key] = Map({})
                    
                target = ensure_validator_dict(case_validator._validator[key])
                case_validator._validator[key] = target
                self.merge_recursive(val, target)
            else:
                if key not in case_validator._validator:
                    case_validator._validator[key] = val

    def rebuild_validator_recursive(self, validator):
        validator = ensure_validator_dict(validator)
        if not hasattr(validator, '_validator') or not isinstance(validator._validator, dict):
            return validator

        new_dict = {}
        for key, val in validator._validator.items():
            val_unpacked = ensure_validator_dict(val)
            if hasattr(val_unpacked, '_validator') and isinstance(val_unpacked._validator, dict):
                new_dict[key] = self.rebuild_validator_recursive(val_unpacked)
            else:
                new_dict[key] = val_unpacked

        if isinstance(validator, MapCombined):
            return MapCombined(new_dict, validator.key_validator, getattr(validator, '_value_validator', None))
        return Map(new_dict)

    def _build(self):
        control_validator = copy.deepcopy(unpack(self.control_validator))
        if self.control_source:
            if isinstance(self.control_source, str):
                self.control_source = [self.control_source]
            for key in reversed(self.control_source):
                map_layer = Map({})
                map_layer._validator[key] = control_validator
                control_validator = map_layer

        case_validator = copy.deepcopy(ensure_validator_dict(self.case_validator))

        if hasattr(case_validator, 'control') and hasattr(case_validator.control, '_validator'):
            case_validator.control._validator = ValidatorBuilder(
                control_validator, case_validator.control._validator
            ).validator
            return case_validator

        self.merge_recursive(control_validator, case_validator)
        final_validator = self.rebuild_validator_recursive(case_validator)

        return final_validator



from strictyaml import Validator
from collections.abc import Callable


class Block:
    def __init__(
        self,
        when: Callable[..., bool],
        schema: Validator,
        constraints: list[Callable[..., bool]] | None = None,
    ):
        assert isinstance(schema, Validator), "schema must be of type Validator"
        self.when = when
        self._validator = schema
        self.constraints = constraints

    def __repr__(self):
        return "{0}(when={1}, schema={2}{3})".format(
            self.__class__.__name__,
            repr(self.when),
            repr(self._validator),
            ", constraints={0}".format(repr(self.constraints)) if self.constraints else "",
        )


class Case(Block):
    def __init__(
        self,
        when: Callable[..., bool],
        schema: Validator,
        constraints: list[Callable[..., bool]] | None = None,
    ):
        super().__init__(when, schema, constraints)


# TODO: implement in DMap
class Overlay(Block):
    def __init__(
        self,
        when: Callable[..., bool],
        schema: Validator,
        constraints: list[Callable[..., bool]] | None = None,
    ):
        super().__init__(when, schema, constraints)
